package org.artifactory.webapp.main;

import com.google.common.collect.Lists;
import org.apache.commons.lang3.SystemUtils;
import org.artifactory.webapp.WebappUtils;
import org.jfrog.common.platform.test.helpers.ProcessUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.zeroturnaround.exec.ProcessExecutor;
import org.zeroturnaround.exec.StartedProcess;

import java.io.File;
import java.io.IOException;
import java.util.List;

import static org.artifactory.webapp.main.AccessProcess.TEST_JFROG_JOIN_KEY;
import static org.jfrog.config.home.JFrogHome.PROP_HOME_DIR;

/**
 * Frontend Server process wrapper. For development use only.
 *
 * @author Yuval Reches
 */
public class FrontendServerProcess {
    private static final Logger log = LoggerFactory.getLogger(FrontendServerProcess.class);

    private File frontendServerHome;
    private StartedProcess process;
    private int port;
    private int routerPort;

    /**
     * @param productHome is the path to <homeDir> (/var should be under it)
     */
    public FrontendServerProcess(File productHome, int routerPort, int port) {
        this.routerPort = routerPort;
        this.port = port;
        if (productHome != null) {
            this.frontendServerHome = productHome;
        } else {
            this.frontendServerHome = new File(System.getProperty("user.home") + "/.frontend");
        }
    }

    FrontendServerProcess(int port, int routerPort, File productHome) {
        this.port = port;
        this.routerPort = routerPort;
        this.frontendServerHome = productHome;
    }

    public void start() throws IOException {
        List<String> cmd = buildCommand();
        port = ProcessUtils.resolvePort(port);
        log.info("Starting external Frontend process with port {}", port);

        File varFolder = new File(frontendServerHome, "var");
        File logs = new File(varFolder, "log/frontend");
        boolean logDirCreated = logs.mkdirs();
        if (!logDirCreated) {
            log.warn("Failed to create frontend logs folder");
        }
        String homePath = frontendServerHome.getAbsolutePath();

        process = new ProcessExecutor()
                .command(cmd)
                .environment(PROP_HOME_DIR, homePath)
                .environment("JF_FRONTEND_PORT", String.valueOf(port))
                //.environment("JF_FRONTEND_LOGGING_APPLICATION_LEVEL", "debug") // for those who wants debug
                .environment("JF_SHARED_SECURITY_JOINKEY",
                        TEST_JFROG_JOIN_KEY) // currently node doesn't read from filesystem due to unknown reason
                .redirectOutput(System.out)
                .redirectError(System.err)
                .destroyOnExit()
                .start();
        log.info("Frontend process started");
    }

    public void stop() {
        if (process != null && process.getProcess().isAlive()) {
            process.getProcess().destroyForcibly();
        }
    }

    private List<String> buildCommand() {
        return Lists.newArrayList(
                WebappUtils.getNodeExecutable(SystemUtils.IS_OS_MAC).getAbsolutePath(),
                WebappUtils.getFrontendExecutable(SystemUtils.IS_OS_MAC).getAbsolutePath());
    }

}