/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.lifecycle.converter;

import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.ha.HaNodeProperties;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.lifecycle.converter.helpers.ConvertersManagerTestBase;
import org.artifactory.lifecycle.converter.helpers.MockServiceConverter;
import org.artifactory.test.TestUtils;
import org.artifactory.version.ArtifactoryVersion;
import org.fest.assertions.Assertions;
import org.jfrog.config.ifc.ConfigurationManager;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Optional;
import java.util.Properties;

import static org.artifactory.version.ArtifactoryVersionProvider.v301;
import static org.artifactory.version.ArtifactoryVersionProvider.v304;
import static org.artifactory.version.ArtifactoryVersionProvider.v310;
import static org.artifactory.version.ArtifactoryVersionProvider.v560;
import static org.artifactory.version.ArtifactoryVersionProvider.v670;
import static org.artifactory.version.ArtifactoryVersionProvider.v680;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertTrue;

/**
 * @author Gidi Shabat
 */
@Test
public class ConvertersManagerTest extends ConvertersManagerTestBase {

    @DataProvider
    public static Object[][] accessEmigratorDataProvider() {
        return new Object[][] {
                {false},
                {true}
        };
    }

    @BeforeMethod
    public void setup() {
        System.setProperty(ConstantValues.devHa.getPropertyName(), "true");
    }

    @AfterMethod
    public void cleanAfterCreateEnv() {
        System.clearProperty(ConstantValues.devHa.getPropertyName());
        Optional.ofNullable(configurationManager).ifPresent(ConfigurationManager::destroy);
        ArtifactoryContextThreadBinder.unbind();
        ArtifactoryHome.unbind();
    }

    /**
     * Convert artifactory which home is version 3.0.1  DBProperties is 3.0.1 and cluster version is 3.0.1
     */
    public void convertAll301() throws IOException {
        File home = TestUtils.createTempDir(getClass());
        createEnvironment(home, v301.get(), v301.get());
        // Make sure that the artifactory local home properties.file has been updated
        assertTrue(isArtifactoryLocalHomePropertiesHasBeenUpdated(home));
        // Make sure that the artifactory db properties has been updated
        assertTrue(isServiceConvertWasRun(home));
        // Make sure that the artifactory.properties exist in home
        assertTrue(isArtifactoryPropertiesHasBeenUpdated(home));
    }

    /**
     * Convert artifactory which home is version 3.0.4  DBProperties is 3.0.4 and cluster version is 3.0.4
     */
    public void convertAll304() throws IOException {
        File home = TestUtils.createTempDir(getClass());
        createEnvironment(home, v304.get(), v304.get());
        // Make sure that the artifactory local home properties.file has been updated
        assertTrue(isArtifactoryLocalHomePropertiesHasBeenUpdated(home));
        // Make sure that the artifactory db properties has been updated
        assertTrue(isServiceConvertWasRun(home));
        // Make sure that the artifactory.properties exist in home
        assertTrue(isArtifactoryPropertiesHasBeenUpdated(home));
    }

    @Test(dataProvider = "accessEmigratorDataProvider")
    public void convert560WithIsInterestedOverride(boolean expected) throws IOException {
        File home = TestUtils.createTempDir(getClass());
        createEnvironment(home, v560.get(), v560.get(), new MockServiceConverter(home + "/.history/db.test", expected));
        assertEquals(isServiceConvertWasRun(home), expected);
    }

    @Test
    public void testConvertAlwaysInterested() throws IOException {
        File home = TestUtils.createTempDir(getClass());
        createEnvironment(home, v670.get(), v680.get(), new MockServiceConverter(home + "/.history/db.test", false));
        assertFalse(isServiceConvertWasRun(home));
        ArtifactoryHome artifactoryHome = new ArtifactoryHome(new File(home + "/.artifactory").getAbsolutePath());
        artifactoryHome.initHaNodeProperties();
        HaNodeProperties haNodeProperties = artifactoryHome.getHaNodeProperties();

        File artifactoryHaNodePropertiesFile = artifactoryHome.getArtifactoryHaNodePropertiesFile();
        Properties properties = new Properties();
        try (FileInputStream inputStream = new FileInputStream(artifactoryHaNodePropertiesFile)) {
            properties.load(inputStream);
        }
        assertNotNull(haNodeProperties);
        Assertions.assertThat(properties).isNotEmpty();
        assertFalse(properties.contains("primary"));
    }

    /**
     * Convert artifactory which home is version current  DBProperties is current and cluster version is current
     */
    public void convertAllCurrent() throws IOException {
        File home = TestUtils.createTempDir(getClass());
        ArtifactoryVersion current = ArtifactoryVersion.getCurrent();
        createEnvironment(home, current, current);
        // Make sure that the artifactory local home properties.file has been updated
        assertFalse(isArtifactoryLocalHomePropertiesHasBeenUpdated(home));
        assertFalse(isArtifactorySharedHomePropertiesHasBeenUpdated(home));
        // Make sure that the artifactory db properties has been updated
        assertFalse(isServiceConvertWasRun(home));
        // Make sure that the artifactory.properties exist in home
        assertTrue(isArtifactoryPropertiesHasBeenUpdated(home));
    }

    /**
     * Convert artifactory which home is version null  DBProperties is null and cluster version is null
     */
    public void convertNoVersions() throws IOException {
        File home = TestUtils.createTempDir(getClass());
        createEnvironment(home, null, null);
        // Make sure that the artifactory local home properties.file has been updated
        assertFalse(isArtifactoryLocalHomePropertiesHasBeenUpdated(home));
        // Make sure that the artifactory db properties has been updated
        assertFalse(isServiceConvertWasRun(home));
        // Make sure that the artifactory.properties exist in home
        assertTrue(isArtifactoryPropertiesHasBeenUpdated(home));
    }

    /**
     * Convert artifactory which home is current  DBProperties is 3.0.1 and cluster version is current
     * <p/>
     * In case that the db_properties doesn't exist then take the version from home.
     * This is dangerous: in case that the home version is current no db conversion will run
     */
    public void convertCurrentHomeDb301Cluster() throws IOException {
        File home = TestUtils.createTempDir(getClass());
        ArtifactoryVersion current = ArtifactoryVersion.getCurrent();
        createEnvironment(home, current, v301.get());
        // Make sure that the artifactory local home properties.file has been updated
        assertFalse(isArtifactoryLocalHomePropertiesHasBeenUpdated(home));
        assertFalse(isArtifactorySharedHomePropertiesHasBeenUpdated(home));
        // Make sure that the artifactory db properties has been updated
        assertTrue(isServiceConvertWasRun(home));
        // Make sure that the artifactory.properties exist in home
        assertTrue(isArtifactoryPropertiesHasBeenUpdated(home));
    }

    /**
     * * Convert artifactory which home is version current  DBProperties is 3.1.0 and cluster version is current
     */
    public void convertCurrentHomeOldDbCurrentCluster() throws IOException {
        File home = TestUtils.createTempDir(getClass());
        ArtifactoryVersion current = ArtifactoryVersion.getCurrent();
        createEnvironment(home, current, v310.get());
        // Make sure that the artifactory local home properties.file has been updated
        assertFalse(isArtifactoryLocalHomePropertiesHasBeenUpdated(home));
        // Make sure that the artifactory db properties has been updated
        assertTrue(isServiceConvertWasRun(home));
        // Make sure that the artifactory.properties exist in home
        assertTrue(isArtifactoryPropertiesHasBeenUpdated(home));
    }

    /**
     * Convert artifactory which home is version 3.0.1  DBProperties is current and cluster version is current
     */
    public void convertOldHomeCurrentDbCurrentCluster() throws IOException {
        File home = TestUtils.createTempDir(getClass());
        ArtifactoryVersion current = ArtifactoryVersion.getCurrent();
        createEnvironment(home, v301.get(), current);
        // Make sure that the artifactory local home properties.file has been updated
        assertTrue(isArtifactoryLocalHomePropertiesHasBeenUpdated(home));
        // Make sure that the artifactory db properties has been updated
        assertFalse(isServiceConvertWasRun(home));
        // Make sure that the artifactory.properties exist in home
        assertTrue(isArtifactoryPropertiesHasBeenUpdated(home));
    }
}






